"""Model the elements picked up from the filesystem"""

import json
import os
from importlib import import_module

import yaml

from django.template import TemplateDoesNotExist
from django.template.loader import select_template
from django.utils.functional import cached_property

from mote import PROJECT_PATHS
from mote.utils import get_object_by_dotted_name


RESERVED_IDS = (
    "mote", "json", "metadata.json", "projects", "aspects", "patterns",
    "elements", "variations"
)


class Base(object):

    def __init__(self):
        if not hasattr(self, "id"):
            self._id = ""

    @property
    def id(self):
        return self._id

    @property
    def relative_path(self):
        """The relative path to the object, eg.
        myproject/website/src/patterns/atoms/button/."""
        raise NotImplementedError

    @property
    def _relative_paths(self):
        """List of paths to search. Used by the layering machinery."""
        li = [self.relative_path]
        part = "/".join(self.relative_path.split("/")[1:])
        for parent_id in reversed(self.project.metadata.get("parents", [])):
            li.append("%s/%s" % (parent_id, part))
        return li

    def _get_template(self, name):
        try:
            return select_template([pth + name for pth in self._relative_paths])
        except TemplateDoesNotExist:
            return None

    @cached_property
    def metadata(self):
        t = None
        try:
            t = select_template([self.relative_path + "metadata.json"])
        except TemplateDoesNotExist:
            pass
        if t is None:
            t = self._get_template("metadata.json")

        if t is not None:
            return json.loads(t.template.source)

        return {}

    @property
    def title(self):
        return self.metadata.get("title", self.id)

    @property
    def description(self):
        return self.metadata.get("description", None)

    @cached_property
    def data(self):
        """Return the data for this object"""
        # Handle all styles of data
        for name in ("data.json", "data.yaml", "json/data.json"):
            t = self._get_template(name)
            if t is not None:
                if name.endswith(".yaml"):
                    return yaml.load(t.template.source)
                else:
                    return json.loads(t.template.source)
        return {}

    def _get_children(self, klass, subdirectory=""):
        """Helper method to fetch children of a certain type"""
        li = []

        # Our own
        t = select_template([self.project.relative_path + "metadata.json"])
        pth = os.path.join(os.path.dirname(t.template.origin.name), "..", self.relative_path, subdirectory)
        if os.path.exists(pth):
            li = [klass(id, self) for id in os.listdir(pth) if not id.startswith(".") and not id in RESERVED_IDS]

        # Ask parents
        attr = klass.__name__.lower() + "s"
        for project_id in reversed(self.project.metadata.get("parents", [])):
            obj = get_object_by_dotted_name(".".join([str(project_id)] + self.dotted_name.split(".")[1:]))
            li.extend(getattr(obj, attr))

        # Remove duplicates
        processed = []
        result = []
        for l in li:
            if l.id not in processed:
                result.append(l)
                processed.append(l.id)

        return sorted(result, key=lambda item: item.metadata.get("position"))


class Variation(Base):
    """A variation *is* an element but the subclassing breaks down"""

    def __init__(self, id, element):
        self._id = id
        self._element = element
        super(Variation, self).__init__()

    @property
    def element(self):
        return self._element

    @property
    def pattern(self):
        return self.element.pattern

    @property
    def aspect(self):
        return self.element.aspect

    @property
    def project(self):
        return self.element.aspect.project

    @property
    def relative_path(self):
        return "%s/%s/src/patterns/%s/%s/variations/%s/" % (
            self.project.id,
            self.aspect.id,
            self.pattern.id,
            self.element.id,
            self.id
        )

    @property
    def template_names(self):
        return [pth + "element.html" for pth in self._relative_paths]

    @property
    def dotted_name(self):
        return "%s.%s.%s.%s.%s" % (
            self.project.id,
            self.aspect.id,
            self.pattern.id,
            self.element.id,
            self.id
        )

    @property
    def modified(self):
        t = self._get_template("element.html")

        if t is not None:
            try:
                return os.path.getmtime(t.template.origin.name)
            except OSError:
                return 0

        return 0


class Element(Base):

    def __init__(self, id, pattern):
        self._id = id
        self._pattern = pattern
        super(Element, self).__init__()

    @property
    def pattern(self):
        return self._pattern

    @property
    def aspect(self):
        return self.pattern.aspect

    @property
    def project(self):
        return self.pattern.aspect.project

    @property
    def relative_path(self):
        return "%s/%s/src/patterns/%s/%s/" % (
            self.project.id,
            self.aspect.id,
            self.pattern.id,
            self.id
        )

    @property
    def dotted_name(self):
        return "%s.%s.%s.%s" % (
            self.project.id,
            self.aspect.id,
            self.pattern.id,
            self.id,
        )

    @property
    def template_names(self):
        return [pth + "element.html" for pth in self._relative_paths]

    @property
    def index_template_names(self):
        return [pth + "index.html" for pth in self._relative_paths]

    @cached_property
    def variations(self):
        return self._get_children(Variation, "variations")

    @property
    def modified(self):
        t = self._get_template("element.html")

        if t is not None:
            try:
                return os.path.getmtime(t.template.origin.name)
            except OSError:
                return 0

        return 0

    def __getattr__(self, key):
        """Allow variation lookup by name"""
        return {e.id: e for e in self.variations}.get(key)


class Pattern(Base):

    def __init__(self, id, aspect):
        self._id = id
        self._aspect = aspect
        super(Pattern, self).__init__()

    @property
    def aspect(self):
        return self._aspect

    @property
    def project(self):
        return self.aspect.project

    @property
    def relative_path(self):
        return "%s/%s/src/patterns/%s/" % (
            self.project.id,
            self.aspect.id,
            self.id
        )

    @property
    def dotted_name(self):
        return "%s.%s.%s" % (
            self.project.id,
            self.aspect.id,
            self.id
        )

    @cached_property
    def elements(self):
        return self._get_children(Element)

    def __getattr__(self, key):
        """Allow element lookup by name"""
        return {e.id: e for e in self.elements}.get(key)


class Aspect(Base):
    """Examples of aspects are website, email templates."""

    def __init__(self, id, project):
        self._id = id
        self._project = project
        super(Aspect, self).__init__()

    @property
    def project(self):
        return self._project

    @property
    def relative_path(self):
        return "%s/%s/" % (
            self.project.id,
            self.id,
        )

    @property
    def dotted_name(self):
        return "%s.%s" % (
            self.project.id,
            self.id
        )

    @cached_property
    def patterns(self):
        return self._get_children(Pattern, "src/patterns")

    def __getattr__(self, key):
        """Allow pattern lookup by name"""
        return {e.id: e for e in self.patterns}.get(key)


class Project(Base):

    def __init__(self, id):
        self._id = id
        self._full_path = os.path.join(PROJECT_PATHS[id], id)
        super(Project, self).__init__()

    @cached_property
    def metadata(self):
        """A project must have a metadata file"""
        try:
            select_template([self.relative_path + "metadata.json"])
        except TemplateDoesNotExist:
            raise RuntimeError(
                "Project %s needs a metadata.json file" % self.id
            )
        return super(Project, self).metadata

    @property
    def project(self):
        return self

    @property
    def relative_path(self):
        return str(self.id) + "/"

    @property
    def dotted_name(self):
        return str(self.id)

    @cached_property
    def aspects(self):
        return self._get_children(Aspect)

    def __getattr__(self, key):
        """Allow aspect lookup by name"""
        return {e.id: e for e in self.aspects}.get(key)



from django import template
from django.test import TestCase
from django.test.client import RequestFactory

from mote import models
from mote.utils import get_object_by_dotted_name


class TagsTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        super(TagsTestCase, cls).setUpTestData()
        cls.factory = RequestFactory()

    def test_render_element_by_identifier(self):
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Lorem ipsum</i>
            </button>"""
        )

    def test_render_element_by_self(self):
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "self.website.atoms.button" %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Lorem ipsum</i>
            </button>"""
        )

    def test_render_element_with_kwargs_variable(self):
        request = self.factory.get("/")
        button = {"Italic": {"text": "Foo"}}
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" data=button %}"""
        )
        result = t.render(template.Context({
            "request": request,
            "button": button
        }))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Foo</i>
            </button>"""
        )

    def test_render_element_with_kwargs_dict(self):
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" data='{"Italic": {"text": "Foo"}}' %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Foo</i>
            </button>"""
        )

    def test_render_element_with_kwargs_variables(self):
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" data='{"Italic": {"text": "{{ foo }}"}}' number=number %}"""
        )
        result = t.render(template.Context({
            "request": request,
            "foo": "Foo",
            "number": 1
        }))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Foo</i>
            </button>"""
        )

    def test_render_other_element(self):
        request = self.factory.get("/")

        # Default
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" %}"""
        )
        result = t.render(template.Context({"request": request}))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Lorem ipsum</i>
            </button>"""
        )

        # Specify an element by dotted name
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" data='{"OtherElement": {"element": "myproject.website.atoms.panel"}}' %}"""
        )
        result = t.render(template.Context({"request": request}))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Lorem ipsum</i>
            </button>
            Panel"""
        )

        # Specify an element by dotted name after relative traversal
        t = template.Template("""{% load mote_tags %}
            {% render_element "myproject.website.atoms.button" data='{"OtherElement": {"element": "{{ element.pattern.panel.dotted_name }}" }}' %}"""
        )
        result = t.render(template.Context({"request": request}))
        self.assertHTMLEqual(
            result,
            """<button class="Button Button--solid Button--yellowButtercup">
            <i>Lorem ipsum</i>
            </button>
            Panel"""
        )

    def test_get_element_data(self):
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% get_element_data "tests/fleet.xml" as fleet %}
            <fleet>
            {% for car in fleet.cars %}
                <car>
                    <brand>{{ car.brand }}</brand>
                    <model>{{ car.model }}</model>
                </car>
            {% endfor %}
            <value>{{ fleet.value }}</value>
            </fleet>"""
        )
        result = t.render(template.Context({
            "request": request,
            "cars": [
                {"brand": "Opel", "model": "Astra"},
                {"brand": "Ford", "model": "Ikon"}
            ]
        }))
        expected = """<fleet>
            <car>
                <brand>Opel</brand>
                <model>Astra</model>
            </car>
            <car>
                <brand>Ford</brand>
                <model>Ikon</model>
            </car>
            <value>100</value>
        </fleet>"""

        self.assertHTMLEqual(result, expected)



import json

from django import template
from django.http import JsonResponse, HttpResponseBadRequest
from django.views.generic.base import View

from rest_framework import serializers, generics
from rest_framework.response import Response

from mote.models import Project, Aspect, Pattern, Element, Variation
from mote.utils import get_object_by_dotted_name


class BaseSerializer(serializers.Serializer):
    rendered = serializers.SerializerMethodField()

    model_class = None

    def get_rendered(self, obj):
        # Let the template tag do the rendering for us
        src = "{% load mote_tags %}{% render_element object data=data %}"
        context = template.Context({
            "object": obj,
            "request": self.context["request"]
        })
        data = self.context.get("data", {})
        if data:
            context["data"] = data
        t = template.Template(src)
        return t.render(context)


class VariationSerializer(BaseSerializer):
    model_class = Variation


class ElementSerializer(BaseSerializer):
    model_class = Element


class BaseDetail(generics.GenericAPIView):
    queryset = []
    permission_classes = []
    model_class = None

    def get(self, request, format=None, *args, **kwargs):
        obj = kwargs.pop("object", None)
        if obj is None:
            obj = self.get_object(**kwargs)
        # Not a fan of modifying kwargs but there really are no side effects
        kwargs["request"] = request
        serializer = globals()[self.model_class.__name__ + "Serializer"](
            obj, context=kwargs
        )
        return Response(serializer.data)


class VariationDetail(BaseDetail):
    model_class = Variation

    def get_object(self, project, aspect, pattern, element, variation):
        project = Project(project)
        aspect = Aspect(aspect, project)
        pattern = Pattern(pattern, aspect)
        element = Element(element, pattern)
        variation = Variation(variation, element)
        return variation


class ElementDetail(BaseDetail):
    model_class = Element

    def get_object(self, project, aspect, pattern, element):
        project = Project(project)
        aspect = Aspect(aspect, project)
        pattern = Pattern(pattern, aspect)
        element = Element(element, pattern)
        return element


class Multiplex(View):
    """Decode multiple API calls from the request and return encoded results."""

    def get(self, request, *args, **kwargs):
        try:
            calls = json.loads(request.GET["calls"])
        except ValueError:
            return HttpResponseBadRequest()

        results = []
        for call in calls:
            obj = get_object_by_dotted_name(call["id"])
            view_klass = globals()[obj.__class__.__name__ + "Detail"]
            view = view_klass().as_view()(
                request=request,
                object=obj,
                data=call["data"],
                format="json"
            )
            rendered = view.render().content
            results.append(json.loads(rendered))

        return JsonResponse({"results": results})



import os
from importlib import import_module

from django.apps import AppConfig
from django.conf import settings

from mote import PROJECT_PATHS


class MoteConfig(AppConfig):
    name = "mote"
    verbose_name = "Mote"

    def ready(self):
        for name in settings.INSTALLED_APPS:
            mod = import_module(name)
            # The test environment unfortunately requires special handling
            if name == "mote.tests":
                pth = os.path.join(os.path.dirname(mod.__file__), "mote", "projects")
            else:
                pth = os.path.join(os.path.dirname(mod.__file__), "..", "mote", "projects")
            if os.path.exists(pth):
                for id in os.listdir(pth):
                    if not id.startswith("."):
                        PROJECT_PATHS[id] = pth



from django.conf import settings
from django.http import HttpResponse
from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.templatetags.static import PrefixNode
from django.utils.functional import cached_property
from django.utils.six.moves.urllib.parse import urljoin
from django.views.generic.base import TemplateView

from mote import PROJECT_PATHS
from mote.models import Project, Aspect, Pattern, Element, Variation


class HomeView(TemplateView):

    template_name = "mote/home.html"

    def get_context_data(self, **kwargs):
        context = super(HomeView, self).get_context_data(**kwargs)
        li = []
        for id, pth in PROJECT_PATHS.items():
            li.append(Project(id))
        context["projects"] = sorted(li, key=lambda item: item.metadata.get("position"))
        return context


class ProjectView(TemplateView):
    """Detail view for a project"""

    template_name = "mote/project.html"

    def get_context_data(self, **kwargs):
        context = super(ProjectView, self).get_context_data(**kwargs)
        context["project"] = Project(kwargs["project"])
        context["__mote_project_id__"] = context["project"].id
        return context


class AspectView(TemplateView):
    """Detail view for an aspect"""

    template_name = "mote/aspect.html"

    def get_context_data(self, **kwargs):
        context = super(AspectView, self).get_context_data(**kwargs)
        project = Project(kwargs["project"])
        context["__mote_project_id__"] = project.id
        context["aspect"] = Aspect(kwargs["aspect"], project)
        return context


class PatternView(TemplateView):
    """Detail view for a pattern"""

    template_name = "mote/pattern.html"

    def get_context_data(self, **kwargs):
        context = super(PatternView, self).get_context_data(**kwargs)
        project = Project(kwargs["project"])
        aspect = Aspect(kwargs["aspect"], project)
        pattern = Pattern(kwargs["pattern"], aspect)
        context["__mote_project_id__"] = project.id
        context["pattern"] = pattern

        # A pattern may have an intro view. First look in the pattern itself,
        # then fall back to normal template resolution.
        template_names = (
            "%s/%s/src/patterns/%s/mote/intro.html" % (
                project.id, aspect.id, pattern.id
            ),
            "mote/pattern/intros/%s.html" % pattern.id,
        )
        intro = None
        for template_name in template_names:
            # todo: dont' render it here, only in the template. Just check that it loads.
            try:
                intro = render_to_string(template_name, {})
                break
            except TemplateDoesNotExist:
                pass
        context["intro"] = intro

        return context


class ElementBaseView(TemplateView):

    @cached_property
    def element(self):
        project = Project(self.kwargs["project"])
        aspect = Aspect(self.kwargs["aspect"], project)
        pattern = Pattern(self.kwargs["pattern"], aspect)
        element = Element(self.kwargs["element"], pattern)
        return element

    def get_context_data(self, **kwargs):
        context = super(ElementBaseView, self).get_context_data(**kwargs)
        context["__mote_project_id__"] = self.element.project.id
        context["element"] = self.element
        context["static_root"] = urljoin(
            PrefixNode.handle_simple("STATIC_URL"),
            self.element.aspect.relative_path
        )
        return context


class ElementIndexView(ElementBaseView):
    """Index view for an element. Provides common UI around an element."""

    def get_template_names(self):
        return self.element.index_template_names


class ElementPartialView(ElementBaseView):
    """Element view with no wrapping html and body tags"""

    def get_template_names(self):
        return self.element.template_names

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class ElementIframeView(ElementBaseView):
    """Element view suitable for rendering in an iframe"""

    def get_template_names(self):
        return [
            "%s/mote/element/iframe.html" % self.element.project.id,
            "mote/element/iframe.html"
        ]


class VariationBaseView(ElementBaseView):

    @cached_property
    def variation(self):
        return Variation(self.kwargs["variation"], self.element)

    def get_context_data(self, **kwargs):
        context = super(VariationBaseView, self).get_context_data(**kwargs)
        # Rename some variables so we can re-use templates
        context["original_element"] = self.element
        context["element"] = self.variation
        return context


class VariationPartialView(VariationBaseView):

    def get_template_names(self):
        return self.variation.template_names

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class VariationIframeView(VariationBaseView):
    """Element view suitable for rendering in an iframe"""

    def get_template_names(self):
        return [
            "%s/mote/element/iframe.html" % self.element.project.id,
            "mote/element/iframe.html"
        ]



import os
import glob
from os.path import expanduser

if "VIRTUAL_ENV" in os.environ:
    BASE_DIR = os.path.join(
        glob.glob(os.environ["VIRTUAL_ENV"] +  "/lib/*/site-packages")[0],
        "mote"
    )
else:
    BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = "SECRET_KEY_PLACEHOLDER"

DEBUG = True

TEMPLATE_DEBUG = True

INSTALLED_APPS = (
    "mote",
    "mote.tests",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.staticfiles",
    "rest_framework",
)

TEMPLATE_CONTEXT_PROCESSORS = [
    "django.contrib.auth.context_processors.auth",
    "django.template.context_processors.debug",
    "django.template.context_processors.i18n",
    "django.template.context_processors.media",
    "django.template.context_processors.static",
    "django.template.context_processors.tz",
    "django.template.context_processors.request",
    "django.contrib.messages.context_processors.messages",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": TEMPLATE_CONTEXT_PROCESSORS,
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "mote.loaders.app_directories.Loader",
                "django.template.loaders.app_directories.Loader",
            ]
        },
    },
]

ROOT_URLCONF = "mote.tests.urls"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

USE_TZ = True

STATIC_URL = "/static/"

MOTE = {"project": lambda request: "myproject"}



from django.conf.urls import url, include

from rest_framework.urlpatterns import format_suffix_patterns

from mote import views, api


app_name = "mote"

urlpatterns = [
    url(r"^$", views.HomeView.as_view(), name="home"),

    url(
        r"^project/partial/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/(?P<variation>[\w-]+)/$",
        views.VariationPartialView.as_view(),
        name="variation-partial"
    ),
    url(
        r"^project/iframe/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/(?P<variation>[\w-]+)/$",
        views.VariationIframeView.as_view(),
        name="variation-iframe"
    ),
    url(
        r"^project/partial/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/$",
        views.ElementPartialView.as_view(),
        name="element-partial"
    ),
    url(
        r"^project/iframe/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/$",
        views.ElementIframeView.as_view(),
        name="element-iframe"
    ),
    url(
        r"^project/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/$",
        views.ElementIndexView.as_view(),
        name="element-index"
    ),
    url(
        r"^project/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/$",
        views.PatternView.as_view(),
        name="pattern"
    ),
    url(
        r"^project/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/$",
        views.AspectView.as_view(),
        name="aspect"
    ),
    url(
        r"^project/(?P<project>[\w-]+)/$",
        views.ProjectView.as_view(),
        name="project"
    ),
]


api_urlpatterns = [
    url(
        r"^api/project/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/(?P<variation>[\w-]+)/$",
        api.VariationDetail.as_view(),
        name="api-variation-detail",
    ),
    url(
        r"^api/project/(?P<project>[\w-]+)/(?P<aspect>[\w-]+)/(?P<pattern>[\w-]+)/(?P<element>[\w-]+)/$",
        api.ElementDetail.as_view(),
        name="api-element-detail",
    ),
    url(
        r"^api/multiplex/$",
        api.Multiplex.as_view(),
        name="api-multiplex"
    ),
]

urlpatterns = urlpatterns + format_suffix_patterns(api_urlpatterns)



import json
import re
import warnings
from hashlib import md5

from bs4 import BeautifulSoup
import xmltodict

from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django import template
from django.template.base import VariableDoesNotExist
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.translation import ugettext as _
from django.utils.functional import Promise
from django.utils.six import string_types, text_type

from mote.utils import deepmerge, deephash, get_object_by_dotted_name
from mote.views import ElementPartialView, VariationPartialView,\
    ElementIndexView


register = template.Library()


# todo: consolidate render_element and render_element_index
@register.tag
def render_element(parser, token):
    """{% render_element element_or_identifier [k=v] [k=v] ... %}"""
    tokens = token.split_contents()
    if len(tokens) < 2:
        raise template.TemplateSyntaxError(
            "render_element element_or_identifier [k=v] [k=v] ... %}"
        )
    di = {}
    for t in tokens[2:]:
        k, v = t.split("=")
        di[k] = v
    return RenderElementNode(tokens[1], **di)


class RenderElementNode(template.Node):

    def __init__(self, element_or_identifier, **kwargs):
        self.element_or_identifier = template.Variable(element_or_identifier)
        self.kwargs = {}
        for k, v in kwargs.items():
            self.kwargs[k] = template.Variable(v)

    def render(self, context):
        # We must import late
        from mote.models import Project, Aspect, Pattern, Element, Variation

        element_or_identifier = self.element_or_identifier.resolve(context)

        # If element_or_identifier is a string convert it
        if isinstance(element_or_identifier, string_types):
            # The "self" project triggers a project lookup. It first checks for
            # a context variable (used internally by the Mote explorer) then
            # for a setting (used when calling render_element over the API).
            if element_or_identifier.startswith("self."):
                project_id = context.get("__mote_project_id__", None)
                if project_id is None:
                    try:
                        value = settings.MOTE["project"]
                    except (AttributeError, KeyError):
                        raise RuntimeError(
                            "Define MOTE[\"project\"] setting for project lookup"
                       )
                    if callable(value):
                        project_id = value(context["request"])
                    else:
                        project_id = value
                obj = get_object_by_dotted_name(
                    element_or_identifier.replace("self.", project_id + ".")
                )
            else:
                obj = get_object_by_dotted_name(element_or_identifier)
            if not isinstance(obj, (Element, Variation)):
                raise template.TemplateSyntaxError(
                    "Invalid identifier %s" % element_or_identifier
                )
        else:
            obj = element_or_identifier

        # Set the object in the context as "element"
        with context.push():
            context["element"] = obj

            # Resolve the kwargs
            resolved = {}
            for k, v in self.kwargs.items():
                try:
                    r = v.resolve(context)
                except VariableDoesNotExist:
                    continue
                if isinstance(r, Promise):
                    r = text_type(r)

                # Strings may be interpreted further
                if isinstance(r, string_types):

                    # Attempt to resolve any variables by rendering
                    t = template.Template(r)
                    raw_struct = t.render(context)

                    # Attempt to convert to JSON
                    try:
                        resolved[k] = json.loads(raw_struct)
                    except ValueError:
                        resolved[k] = r

                else:
                    resolved[k] = r

            # Find the correct view and construct view kwargs
            view_kwargs = dict(
                project=obj.project.id,
                aspect=obj.aspect.id,
                pattern=obj.pattern.id,
            )
            if isinstance(obj, Variation):
                view = VariationPartialView
                view_kwargs.update(dict(
                    element=obj.element.id,
                    variation=obj.id
                ))
            else:
                view = ElementPartialView
                view_kwargs.update(dict(
                    element=obj.id
                ))

			# Compute a cache key before we pop from resolved
            li = [obj.modified, deephash(resolved)]
            li.extend(frozenset(sorted(view_kwargs.items())))
            hashed = md5(
                u":".join([text_type(l) for l in li]).encode("utf-8")
            ).hexdigest()
            cache_key = "render-element-%s" % hashed
            cached = cache.get(cache_key, None)
            if cached is not None:
                return cached

           # Automatically perform masking with the default data
            request = context["request"]
            masked = obj.data
            # Omit top-level key
            if masked:
                masked = masked[list(masked.keys())[0]]
            if "data" in request.GET:
                masked = deepmerge(masked, json.loads(request.GET["data"]))
            elif "data" in resolved:
                masked = deepmerge(masked, resolved.pop("data"))
            context["data"] = masked

            # Construct a final kwargs that includes the context
            final_kwargs = context.flatten()
            del final_kwargs["request"]
            final_kwargs.update(resolved)
            final_kwargs.update(view_kwargs)

            # Call the view. Let any error propagate.
            result = view.as_view()(request, **final_kwargs)

            if isinstance(result, TemplateResponse):
                # The result of a generic view
                result.render()
                html = result.rendered_content
            elif isinstance(result, HttpResponse):
                # Old-school view
                html = result.content

            # Make output beautiful for Chris
            if not settings.DEBUG:
                beauty = BeautifulSoup(html, "html.parser")
                html = beauty.prettify()

            cache.set(cache_key, html, 300)
            return html


@register.tag
def render_element_index(parser, token):
    """{% render_element_index object %}"""
    tokens = token.split_contents()
    if len(tokens) != 2:
        raise template.TemplateSyntaxError(
            "render_element_index object %}"
        )
    return RenderElementIndexNode(tokens[1])


class RenderElementIndexNode(template.Node):

    def __init__(self, obj):
        self.obj = template.Variable(obj)

    def render(self, context):
        obj = self.obj.resolve(context)

        # Call the view. Let any error propagate.
        request = context["request"]
        view_kwargs = dict(
            project=obj.project.id,
            aspect=obj.aspect.id,
            pattern=obj.pattern.id,
            element=obj.id
        )
        result = ElementIndexView.as_view()(request, **view_kwargs)

        if isinstance(result, TemplateResponse):
            # The result of a generic view
            result.render()
            html = result.rendered_content
        elif isinstance(result, HttpResponse):
            # Old-school view
            html = result.content
        return html


@register.tag(name="resolve")
def do_resolve(parser, token):
    """{% resolve var [var] [var] ... %}"""
    tokens = token.split_contents()
    if len(tokens) < 2:
        raise template.TemplateSyntaxError(
            "{% resolve var [var] [var] ... }"
        )
    return ResolveNode(*tokens[1:])


class ResolveNode(template.Node):
    """Return first argument that resolves."""

    def __init__(self, *args):
        warnings.warn(
            "ResolveNode is being deprecated. Refactor your templates.",
            DeprecationWarning
        )
        self.args = []
        for arg in args:
            self.args.append(template.Variable(arg))

    def render(self, context):
        for arg in self.args:
            try:
                r = arg.resolve(context)
            except VariableDoesNotExist:
                continue
            if isinstance(r, Promise):
                r = text_type(r)
            return r
        return ""


@register.tag(name="mask")
def do_mask(parser, token):
    """{% mask var as name %}"""
    tokens = token.split_contents()
    if len(tokens) != 4:
        raise template.TemplateSyntaxError(
            "{% mask var as name }"
        )
    return MaskNode(tokens[1], tokens[3])


class MaskNode(template.Node):
    """Resolve a dictionary and update keys if a similarly named dictionary
    exists in the template context. This allows dictionaries to be updated
    partially, making for cleaner templates and smaller JSON files."""

    def __init__(self, var, name):
        warnings.warn(
            "Masking is now implicit. Refactor your templates.",
            DeprecationWarning
        )
        self.var = template.Variable(var)
        # Clean up the name without resorting to resolving a variable
        self.name = name.strip("'").strip('"')

    def render(self, context):
        var = self.var.resolve(context)
        request = context["request"]
        if self.name in request.GET:
            var = deepmerge(var, json.loads(request.GET[self.name]))
        elif self.name in context:
            var = deepmerge(var, context[self.name])
        context[self.name] = var
        return ""


@register.tag(name="get_element_data")
def do_get_element_data(parser, token):
    """{% get_element_data template_name as name %}"""
    tokens = token.split_contents()
    if len(tokens) != 4:
        raise template.TemplateSyntaxError(
            "{% get_element_data template_name as name }"
        )
    return GetElementDataNode(tokens[1], tokens[3])


class GetElementDataNode(template.Node):
    """Use a template to assemble an XML string, convert it to a dictionary
    and set it in the context."""

    def __init__(self, template_name, name):
        self.template_name = template.Variable(template_name)
        # Clean up the name without resorting to resolving a variable
        self.name = name.strip("'").strip('"')

    def render(self, context):
        template_name = self.template_name.resolve(context)
        di = xmltodict.parse(
            render_to_string(template_name, context=context, request=context["request"]),
            force_list={"hex": True}
        )

        # Discard the root node
        di = di[list(di.keys())[0]]

        # In debug mode use expensive conversion for easy debugging
        if settings.DEBUG:
            context[self.name] = json.loads(json.dumps(di))
        else:
            context[self.name] = di

        return ""



from django.test import TestCase

from mote.utils import deepmerge


class UtilsTestCase(TestCase):

    def test_deepmerge(self):
        source = {
            "one": {"aaa": 1, "bbb": 2},
            "two": [1, 2, 3],
            "three":[
                {"actor": {"name": "Tom", "surname": "Hanks"}},
                {"actor": {"name": "Denzel", "surname": "Washington"}},
            ],
            "four": [
                {"actor": {"name": "Alec", "surname": "Baldwin"}},
                {"actor": {"name": "Brad", "surname": "Pitt"}},
            ],
            "five": [
                {
                    "movie": {
                        "title": "Good Will Hunting",
                        "actors": [
                            {"actor": {"name": "Ben", "surname": "Affleck"}}
                        ]
                    }
                }
            ]
        }
        delta = {
            "one": {"bbb": 777, "ccc": 888},
            "two": [3, 4, 5],
            "three": {"actor": {"name": "Colin"}},
            "four": [
                {"actor": {"name": "Stephen"}},
                {"actor": {"name": "Harrison", "surname": "Ford"}},
                {"actor": {"name": "William"}}
            ],
            "five": [
                {
                    "movie": {
                        "title": "Good Will Hunting",
                        "actors": [
                            {"actor": {"name": "Matt", "surname": "Damon"}},
                            {"actor": {"name": "Casey"}}
                        ]
                    }
                }
            ],
            "infinity": {"a": 0}
        }
        result = deepmerge(source, delta)
        self.assertEqual(
            {
            "one": {"aaa": 1, "bbb": 777, "ccc": 888},
            "two": [3, 4, 5],
            "three":[
                {"actor": {"name": "Colin", "surname": "Hanks"}}
            ],
            "four": [
                {"actor": {"name": "Stephen", "surname": "Baldwin"}},
                {"actor": {"name": "Harrison", "surname": "Ford"}},
                {"actor": {"name": "William", "surname": "Baldwin"}}
            ],
            "five": [
                {
                    "movie": {
                        "title": "Good Will Hunting",
                        "actors": [
                            {"actor": {"name": "Matt", "surname": "Damon"}},
                            {"actor": {"name": "Casey", "surname": "Affleck"}}
                        ]
                    }
                }
            ],
            "infinity": {"a": 0}
            },
            result
        )

    def test_deepmerge_nones(self):
        # If both source and delta are None the result must be None
        source = None
        delta = None
        result = deepmerge(source, delta)
        assert(result, None)

        # If delta is None, we want to retain source
        source = {
            "one": {"aaa": 1, "bbb": 2},
            "two": [1, 2, 3],
            "three":[
                {"actor": {"name": "Tom", "surname": "Hanks"}},
                {"actor": {"name": "Denzel", "surname": "Washington"}},
            ],
            "four": [
                {"actor": {"name": "Alec", "surname": "Baldwin"}},
                {"actor": {"name": "Brad", "surname": "Pitt"}},
            ],
            "five": [
                {
                    "movie": {
                        "title": "Good Will Hunting",
                        "actors": [
                            {"actor": {"name": "Ben", "surname": "Affleck"}}
                        ]
                    }
                }
            ]
        }
        delta = None
        result = deepmerge(source, delta)
        self.assertEqual(result, source, "Result must be == source!")

        # If source is None we want to retain source
        source = None
        delta = {
            "one": {"aaa": 1, "bbb": 2},
            "two": [1, 2, 3],
            "three":[
                {"actor": {"name": "Tom", "surname": "Hanks"}},
                {"actor": {"name": "Denzel", "surname": "Washington"}},
            ],
            "four": [
                {"actor": {"name": "Alec", "surname": "Baldwin"}},
                {"actor": {"name": "Brad", "surname": "Pitt"}},
            ],
            "five": [
                {
                    "movie": {
                        "title": "Good Will Hunting",
                        "actors": [
                            {"actor": {"name": "Ben", "surname": "Affleck"}}
                        ]
                    }
                }
            ]
        }
        result = deepmerge(source, delta)
        self.assertEqual(result, source, "Result must be == None!")

    def test_list_with_nones(self):
        """`None` values are discarded from lists.
        """
        source = {
            "three": [
                {"actor": {"name": "Tom", "surname": "Hanks"}},
                {"actor": {"name": "Denzel", "surname": "Washington"}},
            ],
        }

        delta = {
            "three": [
                {"actor": {"name": "Tom", "surname": "Hanks"}},
                None,
            ],
        }
        result = deepmerge(source, delta)
        self.assertEqual(
            result,
            {"three": [{"actor": {"name": "Tom", "surname": "Hanks"}}]}
        )



import os

from django.template.utils import get_app_template_dirs
from django.template.loaders.filesystem import Loader as FilesystemLoader


class Loader(FilesystemLoader):

    def get_dirs(self):
        # The testing environment does not require upward traversal
        return get_app_template_dirs(os.path.join("..", "mote", "projects")) \
            + get_app_template_dirs(os.path.join("mote", "projects"))



from django.test import TestCase

from mote import models


class ModelsTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        super(ModelsTestCase, cls).setUpTestData()
        cls.project = models.Project("myproject")
        cls.aspect = models.Aspect("website", cls.project)
        cls.pattern = models.Pattern("atoms", cls.aspect)
        cls.element = models.Element("button", cls.pattern)

    def test_project(self):
        self.assertEqual(self.project.id, "myproject")
        self.assertEqual(self.project.title, "My Project")
        self.assertEqual(self.aspect.id, self.project.aspects[0].id)

    def test_aspect(self):
        self.assertEqual(self.aspect.id, "website")
        self.assertEqual(self.aspect.title, "Website")
        self.assertEqual(self.pattern.id, self.aspect.patterns[0].id)

    def test_pattern(self):
        self.assertEqual(self.pattern.id, "atoms")
        self.assertEqual(self.pattern.title, "Atoms")
        self.assertEqual(self.element.id, self.pattern.elements[0].id)

    def test_element(self):
        self.assertEqual(self.element.id, "button")
        self.assertEqual(self.element.title, "Button")



from copy import deepcopy


def _deepmerge(source, delta):
    """Recursive helper"""

    # "source" is the normative structure. Keep it.
    if source is None or delta is None:
        return source

    for key, value in delta.items():

        if isinstance(value, dict):
            if (key in source) and isinstance(source[key], list):
                # We expect a list but didn't get one. Do conversion.
                _deepmerge(source, {key: [value]})
            else:
                node = source.setdefault(key, {})
                _deepmerge(node, value)

        elif (key in source) and isinstance(value, list):
            # Use the zero-th item as an archetype
            el = deepcopy(source[key][0])
            source[key] = []
            for n in value:
                # Remove `None` entries from lists.
                if n is not None:
                    source[key].append(deepcopy(el))
            for n, v in enumerate(value):
                if isinstance(v, dict):
                    _deepmerge(source[key][n], v)
                elif v is not None:
                    source[key][n] = v

        else:
            source[key] = value

    return source


def deepmerge(source, delta):
    """Return a deep merge of two dictionaries"""

    return _deepmerge(deepcopy(source), delta)


# From http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
def deephash(o):
    if isinstance(o, (set, tuple, list)):
	    return tuple([deephash(e) for e in o])

    elif not isinstance(o, dict):
        return hash(o)

    new_o = deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = deephash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


def get_object_by_dotted_name(name):
    """Return object identified by eg. a.b.c.d"""

    # Avoid circular import
    from mote.models import Project, Aspect, Pattern, Element, Variation

    li = name.split(".")
    length = len(li)
    project = Project(li[0])
    if length == 1:
        return project
    aspect = Aspect(li[1], project)
    if length == 2:
        return aspect
    pattern = Pattern(li[2], aspect)
    if length == 3:
        return pattern
    element = Element(li[3], pattern)
    if length == 4:
        return element
    variation = Variation(li[4], element)
    if length == 5:
        return variation
    return None



import os
from setuptools import setup, find_packages


def get_version(fname):
    '''Extracts __version__ from {fname}'''
    import re
    verstrline = open(fname, "rt").read()
    mob = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", verstrline, re.M)
    if mob:
        return mob.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (fname,))


setup(
    name="mote-prk",
    version=get_version('mote/__init__.py'),
    description="Mote - the pattern library framework.",
    long_description = open("README.rst", "r").read() + open("AUTHORS.rst", "r").read() + open("CHANGELOG.rst", "r").read(),
    author="Praekelt Consulting",
    author_email="dev@praekelt.com",
    license="BSD",
    packages=find_packages(),
    zip_safe=True,
    install_requires=[
        "django",
        "beautifulsoup4",
        "six",
        "djangorestframework-jwt",
        "djangorestframework",
        "xmltodict",
        "PyYAML"
    ],
    tests_require=[
        "tox",
    ],
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    include_package_data=True
)



from django import template
from django.test import TestCase
from django.test.client import RequestFactory

from mote import models
from mote.utils import get_object_by_dotted_name


class LayersTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        super(LayersTestCase, cls).setUpTestData()
        cls.factory = RequestFactory()

    def test_render_button(self):
        """We override button element and data"""
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myprojectchild.website.atoms.button" %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result, "Button child"
        )

    def test_render_anchor(self):
        """We override anchor element but not data"""
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myprojectchild.website.atoms.anchor" %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result, "Anchor child"
        )

    def test_render_panel(self):
        """We inherit panel fully"""
        request = self.factory.get("/")
        t = template.Template("""{% load mote_tags %}
            {% render_element "myprojectchild.website.atoms.panel" %}"""
        )
        result = t.render(template.Context({
            "request": request
        }))
        self.assertHTMLEqual(
            result, "Panel"
        )

    def test_metadata_button(self):
        """We override button element and data"""
        obj = get_object_by_dotted_name("myprojectchild.website.atoms.button")
        self.assertEqual(obj.metadata["title"], "MyProjectChild Button")

    def test_metadata_anchor(self):
        """We override anchor element but not data"""
        obj = get_object_by_dotted_name("myprojectchild.website.atoms.anchor")
        self.assertEqual(obj.metadata["title"], "Anchor")

    def test_metadata_panel(self):
        """We inherit panel fully"""
        obj = get_object_by_dotted_name("myprojectchild.website.atoms.panel")
        self.assertEqual(obj.metadata["title"], "Panel")





from django.core.urlresolvers import reverse
from django.test import TestCase

from mote import models


class ViewsTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        super(ViewsTestCase, cls).setUpTestData()
        cls.project = models.Project("myproject")
        cls.aspect = models.Aspect("website", cls.project)
        cls.pattern = models.Pattern("atoms", cls.aspect)
        cls.element = models.Element("button", cls.pattern)

    def test_element_partial(self):
        url = reverse(
            "mote:element-partial",
            args=("myproject", "website", "atoms", "button")
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_element_partial_post_request(self):
        response = self.client.post(
            reverse(
                "mote:element-partial",
                args=("myproject", "website", "atoms", "button")
            )
        )
        self.assertEqual(response.status_code, 200)
